from transforms.base import Transform, TransformError

class Replace(Transform):
    """
    Replace occurence of a pattern by another
    if regexp is set to True, old_pattern is treated as a python regular expression and capture groups references like \\1 can be used in new_pattern
    """
    category = ""
    name = "replace"
    icon = "wxART_TRANSFORM"

    def run(self, data:bytes, old_pattern:bytes=b"old", new_pattern:bytes=b"new", regexp:bool=False):
        if regexp:
            import re
            return re.sub(old_pattern, new_pattern, data)
        else:
            return data.replace(old_pattern, new_pattern)


class FillRepeat(Transform):
    """
    Fill buffer using a repeating pattern
    """
    category = ""
    name = "fill (repeat pattern)"
    icon = "wxART_EDIT"


    def run(self, data:bytes, pattern:bytes=b"\x00"):
        osize = len(data)
        data = pattern * (1 + len(data) // len(pattern))
        return data[:osize]


class FillGenerator(Transform):
    """
    Fill buffer using a generated pattern
    """
    category = ""
    name = "fill (python generator)"
    icon = "wxART_EDIT"

    TEMPLATE = """def fill_pattern_generator():
    # generate an (infinite) sequence
    i = 0
    while True:
        yield i & 0xff
        i += 1
    """

    def run(self, data:bytes, pattern_generator:str=TEMPLATE):
        osize = len(data)
        import itertools
        dic = dict(globals())
        exec(pattern_generator, dic)
        gen = itertools.cycle(dic["fill_pattern_generator"]())
        res = bytearray(osize)
        for i in range(osize):
            res[i] = next(gen)
        return res


class Reverse(Transform):
    """
    Reverse bytes, aka: 00 01 02 03 04 05 -> 05 04 03 02 01 00
    """
    category = ""
    name = "reverse"
    icon = "wxART_NOTE_FIXUP"

    def run(self, data:bytes):
        return data[::-1]


class Skip(Transform):
    """
    skip some bytes, aka: 00 01 02 03 04 05 06 07 08 09 -> 00 02 04 06 08

    step: keep every nth byte
    """
    category = ""
    name = "skip"
    icon = "wxART_REDO"

    def run(self, data:bytes, step:int=2):
        return data[::step]    



class Cut(Transform):
    """
    Cut a data buffer, i.e. remove bytes at the beggining and/or end of the buffer

    remove_at_the_beginning: how many initial bytes to cut
    remove_at_the_end: how many initial bytes to cut
    """
    category = ""
    name = "cut"
    icon = "wxART_CUT"

    def run(self, data:bytes, remove_at_the_beginning:int=0, remove_at_the_end:int=0):
        if remove_at_the_end:
            return data[remove_at_the_beginning:-remove_at_the_end]    
        else:
            return data[remove_at_the_beginning:]    


class ExtractDelimitedTokens(Transform):
    """
    Select (i.e. keep) only the tokens (aka bytes sequences) located between two delimiters

    start_delimiter: the starting delimiter
    end_delimiter: the end delimiter
    keep_delimiters: include the delimiters in the output
    multiline: wether a token can span over multiple lines or not
    """
    category = ""
    name = "tokenize"
    icon = "wxART_CUT"

    def run(self, data:bytes, start_delimiter:bytes=b"\"", end_delimiter:bytes=b"\"", 
            keep_delimiters:bool=False, multiline:bool=False):
        import re
        flags = 0
        if multiline:
            flags = flags | re.S
        r = re.compile(f"{re.escape(start_delimiter.decode('ascii'))}([^{end_delimiter.decode('ascii')}]*){re.escape(end_delimiter.decode('ascii'))}".encode("ascii"), flags)
        res = bytearray()
        for m in r.finditer(data):
            if keep_delimiters:
                res.extend(m.group(0))
            else:
                res.extend(m.group(1))
        return res



class Resize(Transform):
    """
    Resize the buffer. If the new size is bigger than the old one, padding_value will be use to fill the void (repeated as necessary)
    """
    category = ""
    name = "resize"
    icon = "wxART_SELECT"

    def run(self, data:bytes, new_size:int=1, padding_value:bytes=b"\x00"):
        res = data[:new_size]        
        diff = new_size - len(res)
        if diff > 0:
            padding_value = padding_value * (1 + diff // len(padding_value))
            res += padding_value[:diff]
        return res



class Custom(Transform):
    """
    Transform the input buffer using custom python code
    """
    category = ""
    name = "user code (python)"
    icon = "wxART_EDIT"

    TEMPLATE = """def operation(input:bytes):
    # input: the data, as a bytes buffer
    # return value: a bytes or bytearray buffer

    return input
"""

    def run(self, data:bytes, python_code:str=TEMPLATE):
        dic = dict(globals())
        exec(python_code, dic)
        return dic["operation"](data)
